#[macro_use]
pub mod types;
pub mod instr;
pub mod values;

use std::{
    mem::MaybeUninit,
    ops::{Deref, DerefMut},
    string::FromUtf8Error,
};

use num_enum::{TryFromPrimitive, TryFromPrimitiveError};
use thiserror::Error;

#[derive(Error, Debug)]
pub enum Error {
    #[error("Invalid type: {0:#x}")]
    InvalidType(u8),
    #[error("Integer (typed {0}) overflow")]
    IntOverflow(&'static str),
    #[error("Invalid string: {0}")]
    StringParse(
        #[from]
        #[source]
        FromUtf8Error,
    ),
    #[error("Not enough bytes to parse")]
    OutOfRange,
}

impl<Enum> From<TryFromPrimitiveError<Enum>> for Error
where
    Enum: TryFromPrimitive<Primitive = u8>,
{
    fn from(err: TryFromPrimitiveError<Enum>) -> Self {
        Error::InvalidType(err.number)
    }
}

pub trait ParseResult<'a, T>: Into<Result<(T, &'a [u8]), Error>> {
    fn map_data<F, U>(self, map: F) -> Result<(U, &'a [u8]), Error>
    where
        F: FnOnce(T) -> U,
    {
        self.into().map(|(data, input)| (map(data), input))
    }

    fn try_map_data<F, U>(self, try_map: F) -> Result<(U, &'a [u8]), Error>
    where
        F: FnOnce(T) -> Result<U, Error>,
    {
        self.into()
            .and_then(|(data, input)| Ok((try_map(data)?, input)))
    }

    fn or_other<F>(self, other: F) -> Result<(T, &'a [u8]), Error>
    where
        F: FnOnce() -> Result<(T, &'a [u8]), Error>,
    {
        self.into().or_else(|err| match err {
            Error::InvalidType(_) => other(),
            _ => Err(err),
        })
    }
}

impl<'a, T, R: Into<Result<(T, &'a [u8]), Error>>> ParseResult<'a, T> for R {}

pub trait Parse: Sized {
    fn parse(input: &[u8]) -> Result<(Self, &[u8]), Error>;
}

impl<T: Parse, const N: usize> Parse for [T; N] {
    fn parse(input: &[u8]) -> Result<(Self, &[u8]), Error> {
        let mut input = input;

        let mut ret = MaybeUninit::uninit_array();
        for elem in &mut ret {
            let (data, next) = T::parse(input)?;
            elem.write(data);
            input = next;
        }

        // SAFETY: All elements are initialized.
        Ok((unsafe { MaybeUninit::array_assume_init(ret) }, input))
    }
}

impl<T: Parse> Parse for Vec<T> {
    fn parse(input: &[u8]) -> Result<(Self, &[u8]), Error> {
        let (len, input) = u32::parse(input)?;
        (0..len).try_fold(
            (Vec::with_capacity(len as usize), input),
            |(mut vec, input), _| match T::parse(input) {
                Ok((elem, n)) => {
                    vec.push(elem);
                    Ok((vec, n))
                }
                Err(err) => Err(err),
            },
        )
    }
}

#[repr(transparent)]
pub struct Seq<T>(Vec<T>);

impl<T: Parse> Parse for Seq<T> {
    fn parse(input: &[u8]) -> Result<(Self, &[u8]), Error> {
        let mut ret = vec![];
        let mut input = input;
        loop {
            match T::parse(input) {
                Ok((elem, n)) => {
                    ret.push(elem);
                    input = n;
                }
                Err(Error::InvalidType(_)) => break Ok((Seq(ret), input)),
                Err(err) => break Err(err),
            }
        }
    }
}

impl<T> Deref for Seq<T> {
    type Target = Vec<T>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl<T> DerefMut for Seq<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.0
    }
}
