use num_enum::TryFromPrimitive;

use super::{types::ValType, values::TypeIdx, Error, Parse, ParseResult};

pub enum BlockType {
    TypeIdx(TypeIdx),
    ValType(ValType),
    None,
}

impl Parse for BlockType {
    fn parse(input: &[u8]) -> Result<(Self, &[u8]), Error> {
        match ValType::parse(input) {
            Ok((val_type, input)) => Ok((BlockType::ValType(val_type), input)),
            Err(Error::InvalidType(0x40)) => Ok((BlockType::None, &input[1..])),
            Err(Error::InvalidType(_)) => TypeIdx::parse(input).map_data(BlockType::TypeIdx),
            Err(err) => Err(err),
        }
    }
}

types!(
    pub enum Main {
        Unreachable = 0x0,
        Nop = 0x1,
        Block = 0x2,
        Loop = 0x3,
        If = 0x4,
        Else = 0x5,
        End = 0x6,
        Branch = 0xc,
        BranchIf = 0xd,
        BranchTable = 0xe,
        Return = 0xf,
        Call = 0x10,
        CallIndirect = 0x11,

        RefNull = 0xd0,
        RefIsNull = 0xd1,
        RefFunc = 0xd2,

        Drop = 0x1a,
        Select = 0x1b,
        SelectTyped = 0x1c,

        LocalGet = 0x20,
        LocalSet = 0x21,
        LocalTee = 0x22,
        GlobalGet = 0x23,
        GlobalSet = 0x24,

        TableGet = 0x25,
        TableSet = 0x26,

        Sub0xfc = 0xfc,
        Sub0xfd = 0xfd,

        I32Load = 0x28,
        I64Load = 0x29,
        F32Load = 0x2a,
        F64Load = 0x2b,
        I32Load8S = 0x2c,
        I32Load8U = 0x2d,
        I32Load16S = 0x2e,
        I32Load16U = 0x2f,
        I64Load8S = 0x30,
        I64Load8U = 0x31,
        I64Load16S = 0x32,
        I64Load16U = 0x33,
        I64Load32S = 0x34,
        I64Load32U = 0x35,
        I32Store = 0x36,
        I64Store = 0x37,
        F32Store = 0x38,
        F64Store = 0x39,
        I32Store8 = 0x3a,
        I32Store16 = 0x3b,
        I64Store8 = 0x3c,
        I64Store16 = 0x3d,
        I64Store32 = 0x3e,

        MemSize0 = 0x3f,
        MemGrow0 = 0x40,

        I32Const = 0x41,
        I64Const = 0x42,
        F32Const = 0x43,
        F64Const = 0x44,

        I32Eqz = 0x45,
        I32Eq = 0x46,
        I32Ne = 0x47,
        I32LtS = 0x48,
        I32LtU = 0x49,
        I32GtS = 0x4a,
        I32GtU = 0x4b,
        I32LeS = 0x4c,
        I32LeU = 0x4d,
        I32GeS = 0x4e,
        I32GeU = 0x4f,

        I64Eqz = 0x50,
        I64Eq = 0x51,
        I64Ne = 0x52,
        I64LtS = 0x53,
        I64LtU = 0x54,
        I64GtS = 0x55,
        I64GtU = 0x56,
        I64LeS = 0x57,
        I64LeU = 0x58,
        I64GeS = 0x59,
        I64GeU = 0x5a,

        F32Eq = 0x5b,
        F32Ne = 0x5c,
        F32Lt = 0x5d,
        F32Gt = 0x5e,
        F32Le = 0x5f,
        F32Ge = 0x60,

        F64Eq = 0x61,
        F64Ne = 0x62,
        F64Lt = 0x63,
        F64Gt = 0x64,
        F64Le = 0x65,
        F64Ge = 0x66,

        I32Clz = 0x67,
        I32Ctz = 0x68,
        I32PopCnt = 0x69,
        I32Add = 0x6a,
        I32Sub = 0x6b,
        I32Mul = 0x6c,
        I32DivS = 0x6d,
        I32DivU = 0x6e,
        I32RemS = 0x6f,
        I32RemU = 0x70,
        I32And = 0x71,
        I32Or = 0x72,
        I32Xor = 0x73,
        I32Shl = 0x74,
        I32ShrS = 0x75,
        I32ShrU = 0x76,
        I32RotL = 0x77,
        I32RotR = 0x78,

        I64Clz = 0x79,
        I64Ctz = 0x7a,
        I64PopCnt = 0x7b,
        I64Add = 0x7c,
        I64Sub = 0x7d,
        I64Mul = 0x7e,
        I64DivS = 0x7f,
        I64DivU = 0x80,
        I64RemS = 0x81,
        I64RemU = 0x82,
        I64And = 0x83,
        I64Or = 0x84,
        I64Xor = 0x85,
        I64Shl = 0x86,
        I64ShrS = 0x87,
        I64ShrU = 0x88,
        I64RotL = 0x89,
        I64RotR = 0x8a,

        F32Abs = 0x8b,
        F32Neg = 0x8c,
        F32Ceil = 0x8d,
        F32Floor = 0x8e,
        F32Trunc = 0x8f,
        F32Nearest = 0x90,
        F32Sqrt = 0x91,
        F32Add = 0x92,
        F32Sub = 0x93,
        F32Mul = 0x94,
        F32Div = 0x95,
        F32Min = 0x96,
        F32Max = 0x97,
        F32CopySign = 0x98,

        F64Abs = 0x99,
        F64Neg = 0x9a,
        F64Ceil = 0x9b,
        F64Floor = 0x9c,
        F64Trunc = 0x9d,
        F64Nearest = 0x9e,
        F64Sqrt = 0x9f,
        F64Add = 0xa0,
        F64Sub = 0xa1,
        F64Mul = 0xa2,
        F64Div = 0xa3,
        F64Min = 0xa4,
        F64Max = 0xa5,
        F64CopySign = 0xa6,

        I32WrapI64 = 0xa7,
        I32TruncF32S = 0xa8,
        I32TruncF32U = 0xa9,
        I32TruncF64S = 0xaa,
        I32TruncF64U = 0xab,
        I64ExtendI32S = 0xac,
        I64ExtendI32U = 0xad,
        I64TruncF32S = 0xae,
        I64TruncF32U = 0xaf,
        I64TruncF64S = 0xb0,
        I64TruncF64U = 0xb1,
        F32ConvertI32S = 0xb2,
        F32ConvertI32U = 0xb3,
        F32ConvertI64S = 0xb4,
        F32ConvertI64U = 0xb5,
        F32DemoteF64 = 0xb6,
        F64ConvertI32S = 0xb7,
        F64ConvertI32U = 0xb8,
        F64ConvertI64S = 0xb9,
        F64ConvertI64U = 0xba,
        F64PromoteF32 = 0xbb,
        I32ReinterpretF32 = 0xbc,
        I32ReinterpretF64 = 0xbd,
        I64ReinterpretF32 = 0xbe,
        I64ReinterpretF64 = 0xbf,

        I32Extend8S = 0xc0,
        I32Extend16S = 0xc1,
        I64Extend8S = 0xc2,
        I64Extend16S = 0xc3,
        I64Extend32S = 0xc4,
    }

    pub enum Sub0xfc {
        I32TruncSatF32S = 0,
        I32TruncSatF32U = 1,
        I32TruncSatF64S = 2,
        I32TruncSatF64U = 3,
        I64TruncSatF32S = 4,
        I64TruncSatF32U = 5,
        I64TruncSatF64S = 6,
        I64TruncSatF64U = 7,

        MemoryInit = 8,
        DataDrop = 9,
        MemoryCopy = 10,
        MemoryFill = 11,

        TableInit = 12,
        ElemDrop = 13,
        TableCopy = 14,
        TableGrow = 15,
        TableSize = 16,
        TableFill = 17,
    }

    pub enum Sub0xfd {
        V128Load = 0,
        V128Load8x8S = 1,
        V128Load8x8U = 2,
        V128Load16x4S = 3,
        V128Load16x4U = 4,
        V128Load32x2S = 5,
        V128Load32x2U = 6,
        V128Load8Splat = 7,
        V128Load16Splat = 8,
        V128Load32Splat = 9,
        V128Load64Splat = 10,
        V128Load32Zero = 92,
        V128Load64Zero = 93,

        V128Store = 11,
        V128Load8Lane = 84,
        V128Load16Lane = 85,
        V128Load32Lane = 86,
        V128Load64Lane = 87,
        V128Store8Lane = 88,
        V128Store16Lane = 89,
        V128Store32Lane = 90,
        V128Store64Lane = 91,

        V128Const = 12,

        I8x16Shuffle = 13,

        I8x16ExtractLaneS = 21,
        I8x16ExtractLaneU = 22,
        I8x16ReplaceLane = 23,
        I16x8ExtractLaneS = 24,
        I16x8ExtractLaneU = 25,
        I16x8ReplaceLane = 26,
        I32x4ExtractLane = 27,
        I32x4ReplaceLane = 28,
        I64x2ExtractLane = 29,
        I64x2ReplaceLane = 30,
        F32x4ExtractLane = 31,
        F32x4ReplaceLane = 32,
        F64x2ExtractLane = 33,
        F64x2ReplaceLane = 34,

        I8x16Swizzle = 14,
        I8x16Splat = 15,
        I16x8Splat = 16,
        I32x4Splat = 17,
        I64x2Splat = 18,
        F32x4Splat = 19,
        F64x2Splat = 20,

        I8x16Eq = 35,
        I8x16Ne = 36,
        I8x16LtS = 37,
        I8x16LtU = 38,
        I8x16GtS = 39,
        I8x16GtU = 40,
        I8x16LeS = 41,
        I8x16LeU = 42,
        I8x16GeS = 43,
        I8x16GeU = 44,

        I16x8Eq = 45,
        I16x8Ne = 46,
        I16x8LtS = 47,
        I16x8LtU = 48,
        I16x8GtS = 49,
        I16x8GtU = 50,
        I16x8LeS = 51,
        I16x8LeU = 52,
        I16x8GeS = 53,
        I16x8GeU = 54,

        I32x4Eq = 55,
        I32x4Ne = 56,
        I32x4LtS = 57,
        I32x4LtU = 58,
        I32x4GtS = 59,
        I32x4GtU = 60,
        I32x4LeS = 61,
        I32x4LeU = 62,
        I32x4GeS = 63,
        I32x4GeU = 64,

        I64x2Eq = 214,
        I64x2Ne = 215,
        I64x2LtS = 216,
        I64x2GtS = 217,
        I64x2LeS = 218,
        I64x2GeS = 219,

        F32x4Eq = 65,
        F32x4Ne = 66,
        F32x4LtS = 67,
        F32x4GtS = 68,
        F32x4LeS = 69,
        F32x4GeS = 70,

        F64x2Eq = 71,
        F64x2Ne = 72,
        F64x2LtS = 73,
        F64x2GtS = 74,
        F64x2LeS = 75,
        F64x2GeS = 76,

        V128Not = 77,
        V128And = 78,
        V128AndNot = 79,
        V128Or = 80,
        V128Xor = 81,
        V128BitSelect = 82,
        V128AnyTrue = 83,

        I8x16Abs = 96,
        I8x16Neg = 97,
        I8x16PopCnt = 98,
        I8x16AllTrue = 99,
        I8x16Bitmask = 100,
        I8x16NarrowI16x8S = 101,
        I8x16NarrowI16x8U = 102,
        I8x16Shl = 107,
        I8x16ShrS = 108,
        I8x16ShrU = 109,
        I8x16Add = 110,
        I8x16AddSatS = 111,
        I8x16AddSatU = 112,
        I8x16Sub = 113,
        I8x16I8x16SubSatS = 114,
        I8x16SubSatU = 115,
        I8x16MinS = 118,
        I8x16MinU = 119,
        I8x16MaxS = 120,
        I8x16MaxU = 121,
        I8x16AvgrU = 123,

        I16x8ExtAddPairwiseI8x16S = 124,
        I16x8ExtAddPairwiseI8x16U = 125,
        I16x8Abs = 128,
        I16x8Neg = 129,
        I16x8Q15MulrSatS = 130,
        I16x8AllTrue = 131,
        I16x8Bitmask = 132,
        I16x8NarrowI32x4S = 133,
        I16x8NarrowI32x4U = 134,
        I16x8ExtendLowI8x16S = 135,
        I16x8ExtendHighI8x16S = 136,
        I16x8ExtendLowI8x16U = 137,
        I16x8ExtendHighI8x16U = 138,
        I16x8Shl = 139,
        I16x8ShrS = 140,
        I16x8ShrU = 141,
        I16x8Add = 142,
        I16x8AddSatS = 143,
        I16x8AddSatU = 144,
        I16x8Sub = 145,
        I16x8SubSatS = 146,
        I16x8SubSatU = 147,
        I16x8Mul = 149,
        I16x8MinS = 150,
        I16x8MinU = 151,
        I16x8MaxS = 152,
        I16x8MaxU = 153,
        I16x8AvgrU = 154,
        I16x8ExtMulLowI8x16S = 156,
        I16x8ExtMulHighI8x16S = 157,
        I16x8ExtMulLowI8x16U = 158,
        I16x8ExtMulHighI8x16U = 159,

        I32x4ExtAddPairwiseI16x8S = 126,
        I32x4ExtAddPairwiseI16x8U = 127,
        I32x4Abs = 160,
        I32x4Neg = 161,
        I32x4AllTrue = 163,
        I32x4Bitmask = 164,
        I32x4ExtendLowI16x8S = 167,
        I32x4ExtendHighI16x8S = 168,
        I32x4ExtendLowI16x8U = 169,
        I32x4ExtendHighI16x8U = 170,
        I32x4Shl = 171,
        I32x4ShrS = 172,
        I32x4ShrU = 173,
        I32x4Add = 174,
        I32x4Sub = 177,
        I32x4Mul = 181,
        I32x4MinS = 182,
        I32x4MinU = 183,
        I32x4MaxS = 184,
        I32x4MaxU = 185,
        I32x4DotI16x8S = 186,
        I32x4ExtMulLowI16x8S = 188,
        I32x4ExtMulHighI16x8S = 189,
        I32x4ExtMulLowI16x8U = 190,
        I32x4ExtMulHighI16x8U = 191,

        I64x2Abs = 192,
        I64x2Neg = 193,
        I64x2AllTrue = 195,
        I64x2Bitmask = 196,
        I64x2ExtendLowI32x4S = 199,
        I64x2ExtendHighI32x4S = 200,
        I64x2ExtendLowI32x4U = 201,
        I64x2ExtendHighI32x4U = 202,
        I64x2Shl = 203,
        I64x2ShrS = 204,
        I64x2ShrU = 205,
        I64x2Add = 206,
        I64x2Sub = 209,
        I64x2Mul = 213,
        I64x2ExtMulLowI32x4S = 220,
        I64x2ExtMulHighI32x4S = 221,
        I64x2ExtMulLowI32x4U = 222,
        I64x2ExtMulHighI32x4U = 223,

        F32x4Ceil = 103,
        F32x4Floor = 104,
        F32x4Trunc = 105,
        F32x4Nearest = 106,
        F32x4Abs = 224,
        F32x4Neg = 225,
        F32x4Sqrt = 227,
        F32x4Add = 228,
        F32x4Sub = 229,
        F32x4Mul = 230,
        F32x4Div = 231,
        F32x4Min = 232,
        F32x4Max = 233,
        F32x4Pmin = 234,
        F32x4Pmax = 235,

        F64x2Ceil = 116,
        F64x2Floor = 117,
        F64x2Trunc = 122,
        F64x2Nearest = 148,
        F64x2Abs = 236,
        F64x2Neg = 237,
        F64x2Sqrt = 239,
        F64x2Add = 240,
        F64x2Sub = 241,
        F64x2Mul = 242,
        F64x2Div = 243,
        F64x2Min = 244,
        F64x2Max = 245,
        F64x2Pmin = 246,
        F64x2Pmax = 247,

        I32x4TruncSatF32x4S = 248,
        I32x4TruncSatF32x4U = 249,
        F32x4ConvertI32x4S = 250,
        F32x4ConvertI32x4U = 251,
        I32x4TruncSatF64x2SZero = 252,
        I32x4TruncSatF64x2UZero = 253,
        F64x2ConvertLowI32x4S = 254,
        F64x2ConvertLowI32x4U = 255,
        F32x4DemoteF64x2Zero = 94,
        F64x2PromoteLowF32x4 = 95,
    }
);
