use num_enum::TryFromPrimitive;

use super::{Error, Parse, ParseResult};

macro_rules! types {
    ($($v:vis enum $name:ident { $($var:ident = $byte:literal),* $(,)? })+) => {
        $(#[derive(TryFromPrimitive)]
        #[repr(u8)]
        $v enum $name {
            $($var = $byte),*
        }

        impl Parse for $name {
            fn parse(input: &[u8]) -> Result<(Self, &[u8]), Error> {
                u8::parse(input).try_map_data(|data| Ok(Self::try_from_primitive(data)?))
            }
        })+
    };
}

pub struct Limits {
    pub min: u32,
    pub max: u32,
}

impl Parse for Limits {
    fn parse(input: &[u8]) -> Result<(Self, &[u8]), Error> {
        let (sig, input) = u8::parse(input)?;
        match sig {
            0 => u32::parse(input).map_data(|min| Limits { min, max: u32::MAX }),
            1 => u32::parse(input)
                .and_then(|(min, input)| u32::parse(input).map_data(|max| Limits { min, max })),
            _ => Err(Error::InvalidType(sig)),
        }
    }
}

types!(
    pub enum NumType {
        I32 = 0x7f,
        I64 = 0x7e,
        F32 = 0x7d,
        F64 = 0x7c,
    }

    pub enum VecType {
        Vector = 0x7b,
    }

    pub enum RefType {
        Function = 0x70,
        External = 0x6f,
    }
);

pub enum ValType {
    Num(NumType),
    Vec(VecType),
    Ref(RefType),
}

impl Parse for ValType {
    fn parse(input: &[u8]) -> Result<(Self, &[u8]), Error> {
        { NumType::parse(input).map_data(ValType::Num) }
            .or_other(|| VecType::parse(input).map_data(ValType::Vec))
            .or_other(|| RefType::parse(input).map_data(ValType::Ref))
    }
}

pub type ResultType = Vec<ValType>;

pub struct FuncType {
    pub args: ResultType,
    pub returns: ResultType,
}

impl Parse for FuncType {
    fn parse(input: &[u8]) -> Result<(Self, &[u8]), Error> {
        let (sig, input) = u8::parse(input)?;
        match sig {
            0x60 => ResultType::parse(input).and_then(|(args, input)| {
                ResultType::parse(input).map_data(|returns| FuncType { args, returns })
            }),
            _ => Err(Error::InvalidType(sig)),
        }
    }
}

pub type MemType = Limits;

pub struct TableType {
    pub elem_type: RefType,
    pub limits: Limits,
}

impl Parse for TableType {
    fn parse(input: &[u8]) -> Result<(Self, &[u8]), Error> {
        RefType::parse(input).and_then(|(elem_type, input)| {
            Limits::parse(input).map_data(|limits| TableType { elem_type, limits })
        })
    }
}

pub struct GlobalType {
    pub val_type: ValType,
    pub mutability: bool,
}

impl Parse for GlobalType {
    fn parse(input: &[u8]) -> Result<(Self, &[u8]), Error> {
        let (sig, input) = u8::parse(input)?;
        let mutability = match sig {
            0 => false,
            1 => true,
            _ => return Err(Error::InvalidType(sig)),
        };

        let (val_type, input) = ValType::parse(input)?;
        Ok((
            GlobalType {
                val_type,
                mutability,
            },
            input,
        ))
    }
}
