use super::{Error, Parse, ParseResult};

impl Parse for u8 {
    fn parse(input: &[u8]) -> Result<(Self, &[u8]), Error> {
        input
            .split_first()
            .ok_or(Error::OutOfRange)
            .map_data(|&byte| byte)
    }
}

macro_rules! integer {
    (signed: $signed:expr, $ty:ident) => {
        impl Parse for $ty {
            fn parse(input: &[u8]) -> Result<($ty, &[u8]), Error> {
                let mut num = 0;
                let mut shift = 0;
                let mut next = input;
                loop {
                    let (byte, s) = u8::parse(next)?;
                    next = s;
                    num |= ((byte & 0x7f) as $ty)
                        .checked_shl(shift)
                        .filter(|_| shift < (byte & 0x7f).leading_zeros())
                        .ok_or(Error::IntOverflow(stringify!($ty)))?;
                    if byte & 0x80 == 0 {
                        const SIGNED: bool = $signed;
                        if SIGNED && byte & 0x40 != 0 && shift < $ty::BITS {
                            num |= !0 << shift;
                        }
                        break Ok((num, next));
                    }
                    shift += 7;
                }
            }
        }
    };
}

integer!(signed: true, i64);
integer!(signed: true, i32);
integer!(signed: false, u64);
integer!(signed: false, u32);

macro_rules! float {
    ($ty:ident) => {
        impl Parse for $ty {
            fn parse(input: &[u8]) -> Result<($ty, &[u8]), Error> {
                Parse::parse(input).map_data($ty::from_le_bytes)
            }
        }
    };
}

float!(f32);
float!(f64);

impl Parse for String {
    fn parse(input: &[u8]) -> Result<(Self, &[u8]), Error> {
        let (data, next) = Vec::parse(input)?;
        Ok((Self::from_utf8(data)?, next))
    }
}

macro_rules! idx {
    ($name:ident) => {
        pub struct $name(pub u32);

        impl Parse for $name {
            fn parse(input: &[u8]) -> Result<(Self, &[u8]), Error> {
                u32::parse(input).map_data($name)
            }
        }
    };
}

idx!(TypeIdx);
idx!(FuncIdx);
idx!(TableIdx);
idx!(MemIdx);
idx!(GlobalIdx);
idx!(ElemIdx);
idx!(DataIdx);
idx!(LocalIdx);
idx!(LabelIdx);
